import json
import logging
from pprint import pformat
import random
import sys
import click

import matplotlib.pyplot as plt
import mlflow
import mlflow.pyfunc
import mlflow.keras
import mlflow.exceptions

import numpy as np

from skimage.segmentation import mark_boundaries
from sklearn.metrics import roc_auc_score, classification_report
from flass.model import train, get_data, plot_incorrect

logging.basicConfig(
    stream=sys.stdout,
    level=logging.INFO,
    format="%(asctime)s %(name)s %(levelname)-2s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)

try:
    from lime import lime_image
    from lime.wrappers.scikit_image import SegmentationAlgorithm

except ImportError:
    logger.warning("Dependency lime not found, please install with [lime] to enable")


@click.option("--plot/--no-plot", default=False)
@click.option("--batch-size", default=256)
@click.option("--epochs", default=3)
@click.option("--dataset", required=True, help="Choose between {fashion, mnist}")
@click.option("--model-type", required=False, default="kerasconv")
@click.option("--subset", required=False, default=-1)
@click.option("--lime/--no-lime", default=False)
@click.option("--run-name", required=False)
@click.command()
def flass(plot, batch_size, epochs, dataset, model_type, subset, lime, run_name):
    logger.info("Obtaining data")
    data, class_names = get_data(dataset, subset)

    (x_train, y_train), (x_test, y_test) = data

    with mlflow.start_run(run_name=run_name):
        logger.info(f"Artifact URI: {mlflow.get_artifact_uri()}")
        logger.info(f"Tracking URI: {mlflow.get_tracking_uri()}")
        mlflow.log_param("batch_size", batch_size)
        mlflow.log_param("epochs", epochs)
        mlflow.log_param("num_train_instances", len(x_train))
        mlflow.log_param("ml_method", model_type)
        trained_model = train(
            x_train,
            y_train,
            batch_size=batch_size,
            epochs=epochs,
            model_type=model_type,
        )

        trained_model.log_model(trained_model, "saved-model")
        model_location = mlflow.get_artifact_uri("saved-model")
        logger.info(f"Loading model from: {model_location}")

        # The pyfunc flavour of Keras seems to require a Pandas dataframe, but the
        # Keras flavour (as used when using mlflow.keras) appears to accept the higher
        # dimensioned numpy input
        try:
            loaded_model = mlflow.keras.load_model(model_location)
            logger.info("Loaded MLFlow model with keras flavour")
        except mlflow.exceptions.MlflowException:
            loaded_model = mlflow.pyfunc.load_model(model_location)
            logger.info("Loaded MLFlow model with pyfunc flavour")

        if lime:
            # Do a LIME
            samples = random.sample(range(0, len(x_test)), 10)
            for i in samples:
                limeify(x_test[i], loaded_model, class_names)

        predicted_y_probabilities = loaded_model.predict(x_test)
        roc_auc = roc_auc_score(y_test, predicted_y_probabilities, multi_class="ovr")
        mlflow.log_metric("AUC", roc_auc)
        logger.info(f"AUC: {roc_auc}")
        if predicted_y_probabilities.shape[-1] > 1:
            y_predicted = predicted_y_probabilities.argmax(axis=-1)
        else:
            y_predicted = (predicted_y_probabilities > 0.5).astype("int32")

        matching_predictions = (y_predicted == y_test).tolist()
        correct = matching_predictions.count(True)
        incorrect = matching_predictions.count(False)
        mlflow.log_metric("correct_count", correct)
        logger.info(f"Correct predictions: {correct}")
        mlflow.log_metric("incorrect_count", incorrect)
        logger.info(f"Incorrect classifications: {incorrect}")
        logger.info("See MLFlow for more details from run")

        report = classification_report(
            y_test, y_predicted, target_names=class_names, output_dict=True
        )
        classification_report_file = "/tmp/classification_report.json"
        with open(classification_report_file, "w") as file:
            json.dump(report, file)
        logger.info(
            f'Wrote classification report to file "{classification_report_file}"'
        )
        for key in report.keys():
            if type(report[key]) == dict:
                for metric in report[key].keys():
                    mlflow.log_metric(f"{key}-{metric}", report[key][metric])
            else:
                mlflow.log_metric(f"{key}", report[key])

    if plot:
        plot_incorrect(x_test, y_test, y_predicted, class_names)


def limeify(image_to_explain, trained_pipeline, class_names):
    logger.info("Start a LIME explanation")
    lime_image_probabilities = trained_pipeline.predict(np.array([image_to_explain]))[0]
    image_probabilities = tuple(zip(class_names, lime_image_probabilities))
    logger.info(
        "Models predicted probabilities for image:\n" + pformat(image_probabilities)
    )
    plt.imshow(image_to_explain)
    plt.show()
    explainer = lime_image.LimeImageExplainer()
    segmenter = SegmentationAlgorithm(
        "quickshift", kernel_size=1, max_dist=200, ratio=0.2
    )
    explanation = explainer.explain_instance(
        image_to_explain,
        trained_pipeline.predict,
        top_labels=10,
        num_samples=1000,
        segmentation_fn=segmenter,
    )
    logger.info("Done with a LIME")
    temp, mask = explanation.get_image_and_mask(
        explanation.top_labels[0], positive_only=False, num_features=5, hide_rest=False
    )
    plt.imshow(mark_boundaries(temp, mask))
    plt.show()


if __name__ == "__main__":
    flass()



from setuptools import setup

setup(
    name="flass",
    author="Kristian Flikka",
    author_email="kristian.flikka@gmail.com",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=[
        "azure-storage-blob",
        "click",
        "graphviz",
        "matplotlib",
        "mlflow",
        "numpy",
        "pandas",
        "pydot",
        "sklearn",
        "scikit-image",
        "tensorflow",
    ],
    setup_requires=["wheel", "setuptools"],
    extras_require={"lime": ["lime"]},
    description="Train Keras Convolutional Neural Network for image classification",
    long_description="Train Keras Convolutional Neural Network for image classification",
    entry_points={"console_scripts": ["flass=flass.cli:flass"]},
    packages=["flass"],
    license="MIT",
    url="http://flikka.net",
    version="0.1.11",
)



import logging
import os

import matplotlib.pyplot as plt
import mlflow.keras
import mlflow.sklearn
import numpy as np

from skimage.color import gray2rgb
from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn import svm
import tensorflow as tf

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logger = logging.getLogger(__name__)


class ImageFlattener(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        n_samples = len(X)
        return X.reshape((n_samples, -1))


def svm_model():
    model = svm.SVC(probability=True)
    model.predict = model.predict_proba
    return model


def conv_model():
    model = tf.keras.models.Sequential()

    # Convolutional layer with 32 filters. Because 32 is best. Always.
    model.add(
        tf.keras.layers.Conv2D(
            filters=32,
            kernel_size=(3, 3),  # Filter size
            strides=(1, 1),
            padding="valid",
            activation="relu",
            input_shape=(28, 28, 3),  # All important images are 28 pixels
        )
    )

    # Max pooling, all the cool people do it.
    model.add(
        tf.keras.layers.MaxPooling2D(
            pool_size=(2, 2),  # Size feature will be mapped to
            strides=(2, 2),  # How the pool "steps" across the feature
        )
    )

    # Dropout, of course
    model.add(tf.keras.layers.Dropout(rate=0.25))  # Randomly disable 25% of neurons

    # Make 2d output 1d to feed into the classification section
    model.add(tf.keras.layers.Flatten())

    # Final layer, outputting 10 probabilities. Cause all good problems have 10 classes
    model.add(
        tf.keras.layers.Dense(
            units=10,  # Output shape
            activation="softmax",  # Softmax Activation Function
        )
    )

    # Build the model
    model.compile(
        loss=tf.keras.losses.sparse_categorical_crossentropy,
        optimizer=tf.keras.optimizers.Adam(),
        metrics=["accuracy"],
    )

    return model


def get_data(data_key, subset=-1):
    logger.info(f"Using {data_key} dataset, reducing training set to {subset} elements")
    if data_key == "fashion":
        data_classes = get_fashion_data()
    elif data_key == "mnist":
        data_classes = get_mnist_data()
    else:
        raise ValueError(f"Unsupported value for 'data_key': {data_key}")

    ((x_train, y_train), (x_test, y_test)), class_names = data_classes

    x_train = np.array([gray2rgb(grayimg) for grayimg in x_train])
    x_test = np.array([gray2rgb(grayimg) for grayimg in x_test])

    if subset != -1:
        x_train = x_train[:subset]
        y_train = y_train[:subset]

    data_classes = ((x_train, y_train), (x_test, y_test)), class_names
    return data_classes


def get_mnist_data():
    logger.info("Downloading MNIST dataset")
    data = tf.keras.datasets.mnist.load_data()
    logger.info("Done downloading MNIST dataset")
    # Map for human readable class names
    class_names = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
    return data, class_names


def get_fashion_data():
    logger.info("Downloading Fashion dataset")
    data = tf.keras.datasets.fashion_mnist.load_data()
    logger.info("Done downloading Fashion dataset")
    # Map for human readable class names
    class_names = [
        "T-shirt/top",
        "Trouser",
        "Pullover",
        "Dress",
        "Coat",
        "Sandal",
        "Shirt",
        "Sneaker",
        "Bag",
        "Ankle boot",
    ]
    return data, class_names


def train(x, y, batch_size, epochs, model_type):
    logger.info(f"Starting to build {model_type} model")
    if model_type == "kerasconv":
        model = conv_model()
        model.summary()
        model.fit(x, y, batch_size=batch_size, epochs=epochs)
        model.log_model = mlflow.keras.log_model
        return model
    elif model_type == "svm":
        pipeline_steps = [("image_flattener", ImageFlattener())]
        model = svm_model()
        pipeline_steps.append(("model", model))
        full_pipeline = Pipeline(steps=pipeline_steps)
        full_pipeline.fit(x, y)
        full_pipeline.log_model = mlflow.sklearn.log_model
        return full_pipeline


def plot_incorrect(x_test, y_test, y_predicted, class_names):
    incorrect = np.nonzero(y_predicted != y_test)[0]
    # Display the first 16 incorrectly classified images from the test data set
    plt.figure(figsize=(15, 8))

    for j, incorrect in enumerate(incorrect[0:8]):
        plt.subplot(2, 4, j + 1)
        plt.xticks([])
        plt.yticks([])
        plt.imshow(x_test[incorrect].reshape(28, 28), cmap="Reds")
        plt.title("Predicted: {}".format(class_names[y_predicted[incorrect]]))
        plt.xlabel("Actual: {}".format(class_names[y_test[incorrect]]))

    plt.show()
